@@ -54,9 +54,57 @@ SENDMAIL_BINARY = "/usr/sbin/sendmail"
 DISTRO_CODENAME = lsb_release.get_distro_information()['CODENAME']
 DISTRO_ID = lsb_release.get_distro_information()['ID']
 
+# progress information is written here
+PROGRESS_LOG="/var/run/unattended-upgrades.progress"
+
 # set from the sigint signal handler
 SIGNAL_STOP_REQUEST=False
 
+class UnattendedUpgradesCache(apt.Cache):
+    
+    def __init__(self, rootdir, allowed_origins):
+        apt.Cache.__init__(self, rootdir=rootdir)
+        self.allowed_origins = allowed_origins
+        # ensure we update the candidate versions 
+        self.adjust_candidate_versions()
+    def clear(self):
+        apt.Cache.clear(self)
+        # ensure we update the candidate versions 
+        self.adjust_candidate_versions()
+    def adjust_candidate_versions(self):
+        """ Adjust candidate versions to match highest allowed origin
+        
+            This adjusts the origin even if the candidate has a higher
+            version
+        """
+        for pkg in self:
+            # important! this avoids downgrades below
+            if not pkg.is_upgradable:
+                continue
+            # check if we have a version in a allowed origin that is
+            # not the candidate
+            new_cand = None
+            for ver in pkg.versions:
+                if is_allowed_origin(ver, self.allowed_origins):
+                    # leave as soon as we have the highest new candidate
+                    new_cand = ver
+                    break
+            if new_cand and new_cand != pkg.candidate:
+                logging.debug("adjusting candidate version: '%s'" % new_cand)
+                pkg.candidate = new_cand
+
+class LogInstallProgress(apt.progress.base.InstallProgress):
+    """ Install progress that writes to self.LOG
+        (/var/run/unattended-upgrades.progress by default)
+    """
+
+    LOG = PROGRESS_LOG 
+
+    def status_change(self, pkg, percent, status):
+        f=open(self.LOG, "w")
+        f.write(_("Progress: %s %% (%s)") % (percent, pkg))
+        f.close()
+        
 
 class Unlocked:
     """ context manager for unlocking the apt lock while cache.commit()
@@ -79,7 +127,7 @@ def is_dpkg_journal_dirty():
     (similar to debSystem::CheckUpdates)
     """
     d = os.path.dirname(
-        apt_pkg.Config.find_file("Dir::State::status"))+"/updates"
+        apt_pkg.config.find_file("Dir::State::status"))+"/updates"
     for f in os.listdir(d):
         if re.match("[0-9]+", f):
             return True
@@ -165,9 +213,10 @@ def match_whitelist_string(whitelist, origin):
 def upgrade_normal(cache, pkgs_to_upgrade, logfile_dpkg):
     error = None
     res = False
+    iprogress = LogInstallProgress()
     try:
         with Unlocked():
-            res = cache.commit()
+            res = cache.commit(install_progress=iprogress)
     except SystemError,e:
         error = e
     if res:
@@ -208,12 +257,22 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
                 logging.debug("found  partition of size %s (%s)" % (len(changes), changes))
                 smallest_partition = changes
             cache.clear()
+
+        # write progress log information
+        f=open(PROGRESS_LOG+".minimal-steps", "w")
+        percent = len(to_upgrade)/len(pkgs_to_upgrade)*100.0
+        f.write(_("Progress: %s %%: (%s)") % (
+                    percent, ",".join(smallest_partition)))
+        f.close()
+
         # apply changes
         logging.debug("applying set %s" % smallest_partition)
         rewind_cache(cache, [cache[name] for name in smallest_partition])
+
         try:
+            iprogress = LogInstallProgress()
             with Unlocked():
-                res = cache.commit()
+                res = cache.commit(install_progress=iprogress)
                 if not res:
                     raise Exception("cache.commit() returned false")
             cache.open()
@@ -230,10 +289,10 @@ def upgrade_in_minimal_steps(cache, pkgs_to_upgrade, logfile_dpkg=""):
             break
     return True
 
-def is_allowed_origin(pkg, allowed_origins):
-    if not pkg.candidate:
+def is_allowed_origin(ver, allowed_origins):
+    if not ver:
         return False
-    for origin in pkg.candidate.origins:
+    for origin in ver.origins:
         for allowed in allowed_origins:
             if match_whitelist_string(allowed, origin):
                 return True
@@ -247,16 +306,24 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist):
             logging.debug("pkg '%s' now marked delete" % pkg.name)
             return False
         if pkg.marked_install or pkg.marked_upgrade:
-            if not is_allowed_origin(pkg, allowed_origins):
+            if not is_allowed_origin(pkg.candidate, allowed_origins):
                 logging.debug("pkg '%s' not in allowed origin" % pkg.name)
                 return False
             for blacklist_regexp in blacklist:
                 if re.match(blacklist_regexp, pkg.name):
-                    logging.debug("pkg '%s' blacklisted" % pkg.name)
+                    logging.debug("skipping blacklisted package '%s'" % pkg.name)
                     return False
             if pkg._pkg.selected_state == apt_pkg.SELSTATE_HOLD:
                 logging.debug("pkg '%s' is on hold" % pkg.name)
                 return False
+            # check if the package is unsafe to upgrade unattended
+            ignore_require_restart = apt_pkg.config.find_b(
+                "Unattended-Upgrade::IgnoreAppsRequireRestart", False)
+            if (pkg.marked_upgrade and
+                ignore_require_restart == False and
+                pkg.candidate.record.get("Upgrade-Requires") == "app-restart"):
+                logging.debug("pkg '%s' requires app-restart, not safe to upgrade unattended")
+                return False
     return True
 
 def pkgname_from_deb(debfile):
@@ -524,7 +591,8 @@ def main(options, rootdir=""):
         sys.exit(1)
 
     # get a cache
-    cache = apt.Cache(rootdir=rootdir)
+    cache = UnattendedUpgradesCache(rootdir=rootdir, 
+                                    allowed_origins=allowed_origins)
     if cache._depcache.broken_count > 0:
         print _("Cache has broken packages, exiting")
         logging.error(_("Cache has broken packages, exiting"))
@@ -537,11 +605,13 @@ def main(options, rootdir=""):
     pkgs_kept_back = []
     pkgs_auto_removable = set([pkg.name for pkg in cache 
                                if pkg.is_auto_removable])
+
+    # now do the actual upgrade
     for pkg in cache:
         if options.debug and pkg.is_upgradable:
             logging.debug("Checking: %s (%s)" % (pkg.name, map(str, pkg.candidate.origins)))
         if (pkg.is_upgradable and 
-	    is_allowed_origin(pkg,allowed_origins)):
+	    is_allowed_origin(pkg.candidate, allowed_origins)):
             try:
                 pkg.mark_upgrade()
                 if check_changes_for_sanity(cache, allowed_origins,
@@ -595,9 +665,14 @@ def main(options, rootdir=""):
             if not item.is_trusted:
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
             if conffile_prompt(item.destfile):
-                # FIXME: skip package (means to re-run the whole marking again
+                # skip package (means to re-run the whole marking again
                 # and making sure that the package will not be pulled in by
-                # some other package again!
+                # some other package again!)
+                # 
+                # print to stdout to ensure that this message is part of
+                # the cron mail
+                print _("Package '%s' has conffile prompt and needs to be upgraded manually") % pkgname_from_deb(item.destfile)
+                # log to the logfile
                 logging.warning(_("Package '%s' has conffile prompt and needs to be upgraded manually") % pkgname_from_deb(item.destfile))
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
                 pkgs_kept_back.append(pkgname_from_deb(item.destfile))
